(**
===============================================================================================
Name    : bitness

Purpose : Detecting the "bitness" (32/64 bit) of a Windows executable file

Stefan Heymann, Tbingen, Germany
===============================================================================================
Date        Author   Version  Changes
-----------------------------------------------------------------------------------------------
2020-08-12  Heymann  1.0.0    Start
*)

program bitness;

{$APPTYPE CONSOLE}

(*$WARN SYMBOL_PLATFORM OFF *)

uses
  SysUtils,
  Classes;

const
  CAppName    = 'bitness';
  CAppVersion = '1.0.0';


// --- "Bitness" function ---------------------------------------------------------------------

type
  TExeBitness = (ebUnknown, ebNative, eb32, ebItanium, eb64);

const
  CExeBitness_Name : array [TExeBitness] of string = ('?', 'Native', '32-bit', 'IA64', '64-bit');

function GetFileBitness (ExeFilename : string) : TExeBitness;
         // Source: https://stackoverflow.com/questions/495244/how-can-i-test-a-windows-dll-file-to-determine-if-it-is-32-bit-or-64-bit
var
  FS  : TFileStream;
  MZ  : array [0..1] of AnsiChar;
  PEH : int32;
  M   : uint16;
const
  PE_POINTER_OFFSET = 60;
  MACHINE_OFFSET    =  4;
begin
  if not FileExists (ExeFilename) then exit (ebUnknown);
  FS := TFileStream.Create (ExeFilename, fmOpenRead or fmShareDenyNone);
  try
    FS.Seek (0, soBeginning);
    FS.Read (MZ, SizeOf (MZ));
    if (MZ [0] <> 'M') or (MZ [1] <> 'Z') then
      exit (ebUnknown);
    FS.Seek (PE_POINTER_OFFSET, soBeginning);
    FS.Read (PEH, SizeOf (PEH));
    FS.Seek (PEH + MACHINE_OFFSET, soBeginning);
    FS.Read (M, SizeOf (M));
    case M of
      $0000 : Result := ebNative;
      $014C : Result := eb32;
      $0200 : Result := ebItanium;
      $8664 : Result := eb64;
      else    Result := ebUnknown;
      end;
  finally
    FS.Free;
    end;
end;


// --- Command Line Parameters ----------------------------------------------------------------

procedure ReadCmdLine (out Filespec : string; out Recurse : boolean);
var
  i : integer;
  P : string;
begin
  Filespec := '';
  Recurse  := false;
  for i := 1 to ParamCount do begin
    P := ParamStr (i);
    if (Copy (P, 1, 1) = '/') or (Copy (P, 1, 1) = '-') then
      Recurse := AnsiSameText (Copy (P, 2, MaxInt), 'r')
    else
      Filespec := P;
    end;
end;


// --- Help "screen" --------------------------------------------------------------------------

procedure HelpScreen;
begin
  Writeln (CAppName, #32, CAppVersion);
  Writeln;
  Writeln ('Detect whether a Windows executable (.exe/.dll) is 32/64 bit.');
  Writeln;
  Writeln ('Usage:         bitness <filespec> [options]');
  Writeln;
  Writeln ('filespec       File(s) to examine, may contain wildcards');
  Writeln;
  Writeln ('Options:');
  Writeln ('  /r  or  -r   Recurse subdirectories');
end;


// --- Examine Files --------------------------------------------------------------------------

procedure ExamineFiles (Filespec : string; Recurse : boolean);

  procedure ExamineFile (Filename : string);
  begin
    Writeln (Filename, #32#32, CExeBitness_Name [GetFileBitness (Filename)]);
  end;

  procedure ExamineFolder (Folder, Filespec : string; Recurse : boolean);
  var
    SR : TSearchRec;
    F  : string;
  begin
    if FindFirst (Folder + Filespec, faAnyFile, SR) = 0 then
      try
        repeat
          if SR.Attr and faDirectory = 0 then begin
            F := SR.FindData.cFileName;
            ExamineFile (Folder + F);
            end;
        until FindNext (SR) <> 0;
      finally
        FindClose (SR);
        end;

    if Recurse then
      if FindFirst (Folder + '*.*', faDirectory, SR) = 0 then
        try
          repeat
            if (SR.Attr and faDirectory <> 0) and (SR.Name <> '.') and (SR.Name <> '..') then begin
              F := SR.FindData.cFileName;
              ExamineFolder (Folder + F + '\', Filespec, true);
              end;
          until FindNext (SR) <> 0;
        finally
          FindClose (SR);
          end;
  end;

begin
  ExamineFolder (ExtractFilePath (Filespec), ExtractFilename (Filespec), Recurse);
end;


var
  Filespec : string;
  Recurse  : boolean;
begin
  ReadCmdLine (Filespec, Recurse);

  if Filespec = '' then begin
    HelpScreen;
    halt (1);
    end;

  ExamineFiles (Filespec, Recurse);
end.
